
import time
import rclpy
from rclpy.node import Node
from rclpy.node import Node,SetParametersResult,Parameter
from rcl_interfaces.srv import GetParameters
from std_msgs.msg import  String
from typing import List
from haptic_pattern.lifecycle import LifecycleNode
from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
import gc
from interfaces.msg import (
    HapticSignal as HapticSignalMsg,
    ObstacleAvoidanceData,
    ObjectTrackingData,
    ImageMatchingData,
    MapPathFollowingData,
    CloudData,
)

from haptic_pattern.utils import (
    HEIGHT,
    SIGNAL_MAP, 
    SINGAL_INTENSITY_RANGE, 
    MAX_SIGNAL_FREQUENCY,
    MAX_SIGNAL_HALT,
    OBS_DIST_TOO_CLOSE_SCREEN_RATIO,
    OBS_HALT_QUICKEN_MULTIPLIER,
    NEW_SIGALS,
    HapticSignal,
)
from rclpy.qos import QoSPresetProfiles
DATA_QOS = QoSPresetProfiles.SENSOR_DATA.value

class HapticPattern(LifecycleNode):
    def __init__(self):
        super().__init__('haptic_pattern_node')
        self.declare_parameters(
            namespace='',
            parameters=[
                ('algo_namespace', None)
            ]
        )
        self.logger = self.get_logger()
        self.config = None
        
    def on_configure(self):     
        self.logger.info("on_configure() is called")
        self.algo_namespace = self.get_parameter('algo_namespace').value         
        self.intensity = SINGAL_INTENSITY_RANGE[0]
        self.freq = MAX_SIGNAL_FREQUENCY
        self.turn_signal_reset()
        self.obs_enable, self.obj_enable, self.match_enable, self.map_enable = False, False, False, False
        return Transition.TRANSITION_CALLBACK_SUCCESS
        
    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_activate(self):
        self.logger.info("on_activate() is called")
        self.haptic_pub = self.create_publisher(HapticSignalMsg, '/haptic_pattern', 2)

        self.obs_sub = self.create_subscription(ObstacleAvoidanceData, f'{self.algo_namespace}/obs_data', self.obs_turn, DATA_QOS)
        self.obj_sub = self.create_subscription(ObjectTrackingData, f'{self.algo_namespace}/obj_data', self.obj_turn, DATA_QOS)
        self.img_sub = self.create_subscription(ImageMatchingData, f'{self.algo_namespace}/match_data', self.match_turn, DATA_QOS)
        self.map_sub = self.create_subscription(MapPathFollowingData, f'{self.algo_namespace}/path_data', self.map_turn, DATA_QOS)
        self.cld_sub = self.create_subscription(CloudData, f'{self.algo_namespace}/cloud_data', self.cloud_turn, DATA_QOS)
        
        self.algo_node_cli = self.create_client(GetParameters, '/algorithms_node/get_parameters')               
        self.add_on_set_parameters_callback(self.config_update)
        self.last_signal_time = self.get_clock().now()
        rate = 10
        self.timer = self.create_timer(1/rate, self.resolve_rules)
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        self.timer.cancel()
        # self.timer.destroy()
        self.timer = None
        
        self.haptic_pub = None
        self.algo_node_cli = None
        
        # self.obs_sub.destroy()
        # self.obj_sub.destroy()
        # self.img_sub.destroy()
        # self.map_sub.destroy()
        # self.cld_sub.destroy()
        self.obs_sub = None
        self.obj_sub = None
        self.img_sub = None
        self.map_sub = None
        self.cld_sub = None
        
        gc.collect()
        
        return Transition.TRANSITION_CALLBACK_SUCCESS
        
    def config_update(self, params: List[Parameter]):
        self.logger.info(f'Update Haptics: {[str(p.name) for p in params]}')
        for param in params:
            if param.name == 'enable':
                self.enable = param.value
        return SetParametersResult(successful=True)
    
    def turn_signal_reset(self):
        self.halt_time = MAX_SIGNAL_HALT
        # Last turn direction
        self.last_left = True
        # State to control which side to vibrate
        self.next_obs_dir = 0.0
        self.next_obs_dist = HEIGHT
        
        # Determine if the device is aligned with the objects and how much to turn if not
        self.next_obj_dir = 0.0
        self.next_obj_dist = HEIGHT
        self.next_obj_align = True

        # Determine if the device is aligned with image matching and how much to turn if not
        self.match_align = False
        self.match_trip_done = False
        self.match_micro_adjust_dir = 0.0
        self.match_dir = 0.0
        self.match_dist = -1.0
        
        # Determine if the device is aligned with the map direction and how much to turn if not
        self.next_map_dir = 0.0
        self.next_map_align = True
        self.next_map_dist = 0.0
        self.end_of_path = False
        self.next_map_coord = [-1,-1]
    
    # Modify the state
    def obs_turn(self, msg: ObstacleAvoidanceData):
        # APF route turning direction, or dir to avoid obstacle (-1: left, 0:middle, 1: right)
        self.next_obs_dir = msg.obs_dir 
        # the distance to the obstacle
        self.next_obs_dist = msg.obs_dist 
    
    # only use for car detection and avoidance now
    def obj_turn(self, msg : ObjectTrackingData):
        # the direction to the object (-90: left, 0:middle, 90: right)
        self.next_obj_dir = msg.obj_dir
        # the distance to the object
        self.next_obj_dist = msg.obj_dist
        # true if the object direction and the user direction is aligned
        self.next_obj_align = msg.obj_align

    def match_turn(self, msg : ImageMatchingData):
        # true if the matching template direction and the user direction is aligned
        self.match_align = msg.match_align
        # true if the user has done the matching template trip
        self.match_trip_done = msg.match_finished_trip
        # micro adjustment to guide the user to face the template correctly (-90: left, 0:middle, 90: right)
        self.match_micro_adjust_dir = msg.match_micro_adjustment
        
        # the dir to the matching template (-90: left, 0:middle, 90: right)
        self.match_dir =  msg.kf_turning_dir
        # the remaining distance to the next matching template
        self.match_dist = msg.kf_dist
        
    def map_turn(self, msg : MapPathFollowingData):
        # the direction to the next map (N is 0, E is 90, S is 180, W is 270)
        self.next_map_dir =  msg.next_map_dir
        # true if the next map direction and the user direction is aligned
        self.next_map_align = msg.next_map_align
        # true if the user finish the map trip
        self.end_of_path = msg.end_of_path    
        
    def cloud_turn(self, msg):
        pass
    def set_enables(self,response):
        res: GetParameters.Response= response.result()
        self.obs_enable = res.values[0].bool_value
        self.obj_enable = res.values[1].bool_value
        self.match_enable = res.values[2].bool_value
        self.map_enable = res.values[3].bool_value
        
    def resolve_rules(self):
        if ((self.get_clock().now() - self.last_signal_time).nanoseconds * 1e-9) < self.halt_time:
            return
        
        obs_sig_dir = SIGNAL_MAP['empty']
        obs_sig = SIGNAL_MAP['empty']
        match_sig = SIGNAL_MAP['empty']
        obj_sig = SIGNAL_MAP['empty']
        map_sig = SIGNAL_MAP['empty']
        response = self.algo_node_cli.call_async(
            GetParameters.Request(
                names=[
                    'obstacle_avoidance.enable',
                    'object_tracking.enable',
                    'image_path_following.enable',
                    'map_path_following.enable',
                ]
            )
        )
        response.add_done_callback(self.set_enables,)
        obs_enable = self.obs_enable
        obj_enable = self.obj_enable
        match_enable = self.match_enable
        map_enable = self.map_enable

        if obs_enable:
            if self.next_obs_dir == 0:
                obs_sig_dir = NEW_SIGALS['forward'](self.freq)
            elif self.next_obs_dir < 0:
                obs_sig_dir = NEW_SIGALS['dir_left'](self.freq)
                self.halt_time *=OBS_HALT_QUICKEN_MULTIPLIER
            elif self.next_obs_dir > 0:
                obs_sig_dir = NEW_SIGALS['dir_right'](self.freq)
                self.halt_time *=OBS_HALT_QUICKEN_MULTIPLIER

        # if the obstacle is in front, stop immediately
        # car too close rule
        obs_zone = OBS_DIST_TOO_CLOSE_SCREEN_RATIO * HEIGHT
        obs_in_front = min(self.next_obs_dist, self.next_obj_dist) < obs_zone        
        if obs_enable or obj_enable:
            # if obs is close , intensity is bigger
            if obs_in_front:
                obs_intensity = SINGAL_INTENSITY_RANGE[0] + \
                    (SINGAL_INTENSITY_RANGE[1] - SINGAL_INTENSITY_RANGE[0]) * \
                    (obs_zone - min(self.next_obs_dist, self.next_obj_dist) / obs_zone)
                obs_sig =  NEW_SIGALS['obs_stop'](obs_intensity)
                self.halt_time *=OBS_HALT_QUICKEN_MULTIPLIER

        ## TODO: Combine the Image matching and obstacle rules together
        ## rework signal representation in python
        MICRO_ADJUST_THRESHOLD = 30
        # resolve the match rule
        if  match_enable:
            # if we arrived at the end of the image matching trip, stop
            if self.match_trip_done:
                match_sig = SIGNAL_MAP['arrived']
            # if we found the target, 
            # or the remaining distance to the next target is larger than 0
            # go straight
            elif self.match_align or self.match_dist >0.0:
                match_sig = SIGNAL_MAP['forward']
            # if we are doing micro adjustment
            if abs(self.match_micro_adjust_dir) > MICRO_ADJUST_THRESHOLD:
                match_sig = self.dir2sig(self.match_micro_adjust_dir)      

            # do the template dir turn first, instead of the micro adjustment      
            if self.match_dir != 0:
                match_sig = self.dir2sig_o(self.match_dir)

        # resolve the map rule
        if map_enable:
            # if we arrived at the end of the path, stop
            if self.end_of_path: 
                map_sig = SIGNAL_MAP['arrived']
            # if we found the path target, go straight
            elif self.next_map_align:
                map_sig = SIGNAL_MAP['forward']
            else:
                map_sig = self.dir2sig(self.next_map_dir)

        # resolve the priority of the rules

        # image matching propelling mode
        if match_enable:
            obs_sig = SIGNAL_MAP['forward'] if obs_sig == SIGNAL_MAP['empty'] else obs_sig
            if match_sig == SIGNAL_MAP['arrived']:
                sig = match_sig
            elif match_sig == SIGNAL_MAP['forward']: 
                sig = obs_sig
            elif match_sig != SIGNAL_MAP['empty']:
                sig = match_sig
            else:
                sig= SIGNAL_MAP['empty']

            # consistent return arrived signal if we are at the end of the path, for now 
            if map_sig == SIGNAL_MAP['arrived'] :
                sig = map_sig
        else:
            if obs_enable or obj_enable:
                sig = obs_sig_dir
                if obs_sig != SIGNAL_MAP['empty']:
                    sig = [obs_sig_dir , obs_sig]
            else:
                sig = SIGNAL_MAP['empty']
                
        if isinstance(sig, HapticSignal) or all(isinstance(s, HapticSignal) for s in sig):
            if isinstance(sig, HapticSignal):
                self.haptic_pub.publish(sig.msg())
            else:
                for s in sig:
                    self.haptic_pub.publish(s.msg())
            self.last_signal_time = self.get_clock().now()
        self.turn_signal_reset()
        
    @staticmethod
    def dir2sig(dir):
        if dir < 0 and dir >= - 90:
            return SIGNAL_MAP['target'](0)
        elif dir > 0 and dir <= 90:
            return SIGNAL_MAP['target'](3)
        elif dir > 90 and dir <= 180:
            return SIGNAL_MAP['target'](2)
        elif dir < -90 and dir >= -180:
            return SIGNAL_MAP['target'](1)
    
    def dir2sig_o(self,dir):
        if dir < 0:
            return [
                HapticSignal(SIGNAL_MAP['stop'](0)),
                HapticSignal(SIGNAL_MAP['stop'](1)),
            ]
        elif dir > 0:
            return [
                HapticSignal(SIGNAL_MAP['stop'](0)),
                HapticSignal(SIGNAL_MAP['stop'](3)),
            ]
            
def main(args=None):
    rclpy.init(args=args)
    node = HapticPattern()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    # node.destroy_node()
    rclpy.shutdown()